package core

import (
	"git.akyoto.dev/cli/q/src/asm"
	"git.akyoto.dev/cli/q/src/ast"
	"git.akyoto.dev/cli/q/src/cpu"
	"git.akyoto.dev/cli/q/src/errors"
	"git.akyoto.dev/cli/q/src/expression"
	"git.akyoto.dev/cli/q/src/token"
	"git.akyoto.dev/cli/q/src/types"
)

// ExpressionToRegister puts the result of an expression into the specified register.
func (f *Function) ExpressionToRegister(node *expression.Expression, register cpu.Register) (types.Type, error) {
	f.SaveRegister(register)

	if node.IsFolded {
		f.RegisterNumber(asm.MOVE, register, node.Value)
		return types.Int, nil
	}

	if node.IsLeaf() {
		return f.TokenToRegister(node.Token, register)
	}

	if ast.IsFunctionCall(node) {
		fn, err := f.CompileCall(node)

		if register != f.CPU.Output[0] {
			f.RegisterRegister(asm.MOVE, register, f.CPU.Output[0])
		}

		if fn == nil || len(fn.Output) == 0 {
			return nil, err
		}

		return fn.Output[0].Type, err
	}

	if node.Token.Kind == token.Array {
		array := f.VariableByName(node.Children[0].Token.Text(f.File.Bytes))
		offset, err := f.Number(node.Children[1].Token)
		f.MemoryRegister(asm.LOAD, asm.Memory{Base: array.Register, Offset: int8(offset), Length: 1}, register)
		return types.Int, err
	}

	if node.Token.Kind == token.Period {
		left := node.Children[0]
		name := left.Token.Text(f.File.Bytes)
		variable := f.VariableByName(name)
		right := node.Children[1]
		name = right.Token.Text(f.File.Bytes)
		field := variable.Type.(*types.Pointer).To.(*types.Struct).FieldByName(name)
		f.MemoryRegister(asm.LOAD, asm.Memory{Base: variable.Register, Offset: int8(field.Offset), Length: byte(field.Type.Size())}, register)
		return field.Type, nil
	}

	if len(node.Children) == 1 {
		if !node.Token.IsUnaryOperator() {
			return nil, errors.New(errors.MissingOperand, f.File, node.Token.End())
		}

		typ, err := f.ExpressionToRegister(node.Children[0], register)

		if err != nil {
			return typ, err
		}

		return typ, f.ExecuteRegister(node.Token, register)
	}

	left := node.Children[0]
	right := node.Children[1]
	final := register

	if f.UsesRegister(right, register) {
		register = f.NewRegister()
	}

	typ, err := f.ExpressionToRegister(left, register)

	if err != nil {
		return nil, err
	}

	if typ == types.AnyPointer && right.Token.Kind == token.Identifier && f.VariableByName(right.Token.Text(f.File.Bytes)).Type == types.AnyPointer {
		typ = types.Int
	}

	err = f.Execute(node.Token, register, right)

	if register != final {
		f.RegisterRegister(asm.MOVE, final, register)
		f.FreeRegister(register)
	}

	return typ, err
}
